import typing as t

import pydantic

import nimbo.core.config.yaml_loader
from nimbo.core.config.aws_config import AwsConfig
from nimbo.core.config.common_config import CloudProvider
from nimbo.core.config.common_config import RequiredCase
from nimbo.core.config.gcp_config import GcpConfig


# noinspection PyUnresolvedReferences
def make_config(config_path: str) -> t.Union[AwsConfig, GcpConfig]:
    config = yaml_loader.from_file(config_path)
    config["config_path"] = config_path

    # Provider field validation is postponed. If cloud_provider is not specified,
    # assume AwsConfig for running commands like --help and generate-config.
    if "cloud_provider" not in config:
        return AwsConfig(**config)
    else:
        config["cloud_provider"] = config["cloud_provider"].upper()

    try:
        cloud_provider = CloudProvider(config["cloud_provider"])
    except ValueError:
        permitted_values = ", ".join([f"'{p.value}'" for p in CloudProvider])
        raise pydantic.ValidationError(
            [
                pydantic.error_wrappers.ErrorWrapper(
                    Exception(
                        f"value is not a valid enumeration member; permitted: "
                        f"{permitted_values}"
                    ),
                    "cloud_provider",
                )
            ],
            AwsConfig,
        )

    if cloud_provider == CloudProvider.AWS:
        return AwsConfig(**config)

    return GcpConfig(**config)



import abc
import os
import socket
import subprocess
import time
from typing import Dict

from nimbo import CONFIG
from nimbo.core.constants import NIMBO_ROOT
from nimbo.core.print import nprint, nprint_header


class Instance(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def run(job_cmd: str, dry_run=False) -> Dict[str, str]:
        ...

    @staticmethod
    @abc.abstractmethod
    def run_access_test(dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def _block_until_instance_running(instance_id: str) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def _get_host_from_instance_id(instance_id: str, dry_run=False) -> str:
        ...

    @staticmethod
    @abc.abstractmethod
    def stop_instance(instance_id: str, dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def resume_instance(instance_id: str, dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def delete_instance(instance_id: str, dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def delete_all_instances(dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def get_status(instance_id: str, dry_run=False) -> str:
        ...

    @staticmethod
    @abc.abstractmethod
    def ls_active_instances(dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def ls_stopped_instances(dry_run=False) -> None:
        ...

    @classmethod
    def ssh(cls, instance_id: str, dry_run=False) -> None:
        host = cls._get_host_from_instance_id(instance_id, dry_run)

        if dry_run:
            return

        subprocess.Popen(
            f"ssh -i {CONFIG.instance_key} "
            f"-o 'StrictHostKeyChecking no' -o ServerAliveInterval=20 "
            f"ubuntu@{host}",
            shell=True,
        ).communicate()

    @classmethod
    def sync_notebooks(cls, instance_id: str):
        host = cls._get_host_from_instance_id(instance_id)

        subprocess.Popen(
            f"rsync -avm -e 'ssh -i {CONFIG.instance_key}' "
            f"--include '*/' --include '*.ipynb' --exclude '*' "
            f"ubuntu@{host}:/home/ubuntu/project/ .",
            shell=True,
        ).communicate()

    @staticmethod
    def _sync_code(host: str) -> None:
        if ".git" not in os.listdir():
            nprint(
                "No git repo found. Syncing all python and bash files as a fallback.",
                style="warning",
            )
            nprint(
                "Please consider using git to track the files to sync.", style="warning"
            )
            subprocess.Popen(
                f"rsync -avm -e 'ssh -i {CONFIG.instance_key}' "
                f"--include '*/' --include '*.py' --include '*.ipynb' --include '*.sh' "
                f"--exclude '*' "
                f". ubuntu@{host}:/home/ubuntu/project",
                shell=True,
            ).communicate()
        else:
            output, error = subprocess.Popen(
                "git ls-tree -r HEAD --name-only",
                stdout=subprocess.PIPE,
                shell=True,
            ).communicate()
            git_tracked_files = output.decode("utf-8").strip().splitlines()
            include_files = [
                f"--include '{file_name}'" for file_name in git_tracked_files
            ]
            include_string = " ".join(include_files)
            subprocess.Popen(
                f"rsync -amr -e 'ssh -i {CONFIG.instance_key}' "
                f"--include '*/' {include_string} --exclude '*' "
                f". ubuntu@{host}:/home/ubuntu/project",
                shell=True,
            ).communicate()

    @staticmethod
    def _block_until_ssh_ready(host: str) -> None:
        nprint_header(
            f"Waiting for instance to be ready for ssh at {host}. "
            "This can take up to 2 minutes... "
        )

        start = time.monotonic()

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)

        reconnect_count = 0
        while reconnect_count < CONFIG.ssh_timeout:
            error_num = sock.connect_ex((host, 22))

            if error_num == 0:
                break

            time.sleep(1)
            reconnect_count += 1
        else:
            raise RuntimeError(
                "Something went wrong while connecting to the instance.\n"
                "Please verify your security groups, instance key and "
                "instance profile, and try again.\n"
                "More info at docs.nimbo.sh/common-issues#cant-ssh.\n"
            )

        nprint_header(f"Ready. (%0.3f s)" % (time.monotonic() - start))

    @staticmethod
    def _run_remote_script(
        ssh_cmd: str,
        scp_cmd: str,
        host: str,
        instance_id: str,
        job_cmd: str,
        script: str,
    ) -> None:

        remote_script = os.path.join(NIMBO_ROOT, "scripts", script)
        subprocess.check_output(
            f"{scp_cmd} {remote_script} ubuntu@{host}:/home/ubuntu/", shell=True
        )

        nimbo_log = "/home/ubuntu/nimbo-log.txt"
        bash_cmd = f"bash {script}"
        if CONFIG.run_in_background:
            full_command = (
                f"nohup {bash_cmd} {instance_id} {job_cmd}"
                f" </dev/null >{nimbo_log} 2>&1 &"
            )
        else:
            full_command = f"{bash_cmd} {instance_id} {job_cmd}"

        subprocess.Popen(
            f'{ssh_cmd} ubuntu@{host} "{full_command}"', shell=True
        ).communicate()



import os

import click

from nimbo import assert_required_config, set_config, cloud_context, IS_TEST_ENV
from nimbo.core.click_extensions import (
    HelpSection,
    NimboCommand,
    NimboGroup,
    pprint_errors,
)
from nimbo.core import utils
from nimbo.core.config import RequiredCase, make_config

_CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], max_content_width=90)
_CONFIG_PATH_OVERRIDE = (
    os.environ["NIMBO_CONFIG"] if "NIMBO_CONFIG" in os.environ else "nimbo-config.yml"
)


@click.group(cls=NimboGroup, context_settings=_CONTEXT_SETTINGS)
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    default=_CONFIG_PATH_OVERRIDE,
    show_default=True,
)
def cli(config):
    """
    Run compute jobs on AWS as if you were running them locally.

    Visit https://docs.nimbo.sh for more help.
    """

    if not IS_TEST_ENV:
        global _CONFIG_PATH_OVERRIDE
        _CONFIG_PATH_OVERRIDE = config

        set_config(config_factory=make_config, config_path=config)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.argument("job_cmd")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.JOB)
@pprint_errors
@cloud_context
def run(cloud, job_cmd, dry_run):
    """Run JOB_CMD on an instance.

    JOB_CMD is any command you would run locally.
    E.g. \"python runner.py --epochs=10\".\n
    The command must be between quotes.
    """
    cloud.run(job_cmd, dry_run)


@cli.command(
    cls=NimboCommand,
    help_section=HelpSection.INSTANCE,
    short_help="Launch Jupiter with your code, data, and environment",
)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.JOB)
@pprint_errors
@cloud_context
def notebook(cloud, dry_run):
    """
    Launch Jupyter Lab on an instance with your code, data and environment.

    Make sure to run 'nimbo sync-notebooks <instance_id>' frequently to sync
    the notebook to your local folder, as the remote notebooks will be lost
    once the instance is terminated.
    """
    cloud.run("_nimbo_notebook", dry_run)


@cli.command(
    cls=NimboCommand,
    help_section=HelpSection.INSTANCE,
    short_help="Launch an instance with minimal setup.",
)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.JOB)
@pprint_errors
@cloud_context
def launch(cloud, dry_run):
    """
    Launch an instance according to your Nimbo config with minimal setup.

    The launched instance does not include your code, data, or environment.
    """
    cloud.run("_nimbo_launch", dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.JOB)
@pprint_errors
@cloud_context
def launch_and_setup(cloud, dry_run):
    """
    Launch an instance with your code, data and environment.

    The launched instance does not run any job.
    """
    cloud.run("_nimbo_launch_and_setup", dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.INSTANCE)
@pprint_errors
@cloud_context
def ssh(cloud, instance_id, dry_run):
    """SSH into an instance by INSTANCE_ID."""
    cloud.ssh(instance_id, dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def get_status(cloud, instance_id, dry_run):
    """Get the status of an instance by INSTANCE_ID."""
    print(cloud.get_status(instance_id, dry_run))


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def ls_active(cloud, dry_run):
    """List all your active instances."""
    cloud.ls_active_instances(dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def ls_stopped(cloud, dry_run):
    """List all your stopped instances."""
    cloud.ls_stopped_instances(dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def rm_instance(cloud, instance_id, dry_run):
    """Terminate an instance by INSTANCE_ID."""
    cloud.delete_instance(instance_id, dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def rm_all_instances(cloud, dry_run):
    """Terminate all your instances."""
    click.confirm(
        "This will delete all your running instances.\n" "Do you want to continue?",
        abort=True,
    )
    cloud.delete_all_instances(dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def stop_instance(cloud, instance_id, dry_run):
    """Stop an instance by INSTANCE_ID."""
    cloud.stop_instance(instance_id, dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.INSTANCE)
@click.argument("instance_id")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def resume_instance(cloud, instance_id, dry_run):
    """Resume a stopped instance by INSTANCE_ID."""
    cloud.resume_instance(instance_id, dry_run)


@cli.command(
    cls=NimboCommand,
    help_section=HelpSection.INSTANCE,
    short_help="Add your IP to instance firewall ingress allow list.",
)
@click.argument("security_group")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def add_current_ip(cloud, security_group, dry_run):
    """Add the IP of the current machine to the allowed inbound rules of GROUP.

    GROUP is the security group to which the inbound rule will be added.
    """
    cloud.allow_ingress_current_ip(security_group, dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.STORAGE)
@click.argument("bucket_name")
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def mk_bucket(cloud, bucket_name, dry_run):
    """
    Create the bucket BUCKET_NAME in S3.

    BUCKET_NAME is the name of the bucket to create, s3://BUCKET_NAME
    """
    cloud.mk_bucket(bucket_name, dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.STORAGE)
@click.argument("path")
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def ls_bucket(cloud, path):
    """List S3 objects in PATH.

    PATH is an S3 path of the form s3://bucket-name/my/files/path.
    """
    cloud.ls_bucket(path)


@cli.command(cls=NimboCommand, help_section=HelpSection.STORAGE)
@click.argument(
    "folder", type=click.Choice(["datasets", "results", "logs"]), required=True
)
@click.option(
    "--delete",
    is_flag=True,
    help="""
      Deletes any files that exist in the local folder
      but don't exist in the remote folder.
    """,
)
@assert_required_config(RequiredCase.STORAGE)
@pprint_errors
@cloud_context
def push(cloud, folder, delete):
    """Push your local datasets/results folder onto S3."""

    if delete:
        click.confirm(
            "This will delete any files that exist in the remote "
            "folder but do not exist in the local folder.\n"
            "Do you want to continue?",
            abort=True,
        )
    cloud.push(folder, delete)


@cli.command(cls=NimboCommand, help_section=HelpSection.STORAGE)
@click.argument(
    "folder", type=click.Choice(["datasets", "results", "logs"]), required=True
)
@click.option(
    "--delete",
    is_flag=True,
    help="""
      Deletes any files that exist in the local
      folder but don't exist in the remote folder.
    """,
)
@assert_required_config(RequiredCase.STORAGE)
@pprint_errors
@cloud_context
def pull(cloud, folder, delete):
    """Pull datasets/results folder into your computer from S3."""

    if delete:
        click.confirm(
            "This will delete any files that exist in the local "
            "folder but do not exist in the remote folder.\n"
            "Do you want to continue?",
            abort=True,
        )
    cloud.pull(folder, delete)


@cli.command(cls=NimboCommand, help_section=HelpSection.STORAGE)
@click.argument("instance_id")
@assert_required_config(RequiredCase.INSTANCE)
@pprint_errors
@cloud_context
def sync_notebooks(cloud, instance_id):
    """
    Pull ipynb files from INSTANCE_ID to your local folder.

    Make sure to run 'nimbo sync-notebooks <instance_id>' to sync the notebook
    to your local folder, as the remote notebooks will be lost once the instance
    is terminated.
    """
    cloud.sync_notebooks(instance_id)


@cli.command(cls=NimboCommand, help_section=HelpSection.UTILS)
@assert_required_config(RequiredCase.NONE)
@pprint_errors
def generate_config():
    """Create a base configuration file in the current directory.

    Remember to change any fields to your own values.
    """
    utils.generate_config(_CONFIG_PATH_OVERRIDE)


@cli.command(cls=NimboCommand, help_section=HelpSection.UTILS)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def mk_instance_key(cloud):
    """Create and download an instance key to the current directory."""
    cloud.mk_instance_key()


@cli.command(cls=NimboCommand, help_section=HelpSection.UTILS)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.INSTANCE, RequiredCase.STORAGE)
@pprint_errors
@cloud_context
def test_access(cloud, dry_run):
    """Run a mock job to test your config."""
    cloud.run_access_test(dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.UTILS)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def ls_prices(cloud, dry_run):
    """List the prices, types, and specs of GPU instances."""
    cloud.ls_gpu_prices(dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.UTILS)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def ls_spot_prices(cloud, dry_run):
    """List the prices, types, and specs of GPU spot instances."""
    cloud.ls_spot_gpu_prices(dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.UTILS)
@click.argument("qty", type=int, required=True)
@click.argument("timescale", type=click.Choice(["days", "months"]), required=True)
@click.option("--dry-run", is_flag=True)
@assert_required_config(RequiredCase.MINIMAL)
@pprint_errors
@cloud_context
def spending(cloud, qty, timescale, dry_run):
    """Show daily/monthly spending summary. Costs without credits or refunds applied.

    QTY is the number of days/months you want to see, starting from the current date.\n
    For example:\n
        'nimbo spending 10 days' shows daily spending of the last 10 days\n
        'nimbo spending 3 months' shows the monthly spending of the last 3 months
    """
    cloud.spending(qty, timescale, dry_run)


@cli.command(cls=NimboCommand, help_section=HelpSection.ADMIN)
@click.argument("profile")
@click.option(
    "--no-s3-access", help="Create your own S3 access role separately.", is_flag=True
)
@assert_required_config(RequiredCase.NONE)
@pprint_errors
@cloud_context
def admin_setup(cloud, profile, no_s3_access):
    """
    Setup Nimbo access role for your organisation.

    Creates a user group and instance role that gives users in your AWS account the
    necessary permissions to use Nimbo. Once `admin-setup` is done, you can run
    `add-user` to allow a specific user to use the user group and role.

    PROFILE is the profile name of your root/admin account from ~/.aws/credentials
    """
    cloud.setup(profile, no_s3_access)


@cli.command(cls=NimboCommand, help_section=HelpSection.ADMIN)
@click.argument("profile")
@click.argument("username")
@assert_required_config(RequiredCase.NONE)
@pprint_errors
@cloud_context
def add_user(cloud, profile, username):
    """Adds user USERNAME to the user group NimboUserGroup.

    You must have run 'nimbo admin-setup' before adding users.

    PROFILE is the profile name of your root/admin account from ~/.aws/credentials
    """
    cloud.add_user(profile, username)



import abc


class Utils(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def ls_gpu_prices(dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def ls_spot_gpu_prices(dry_run=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def spending(qty: int, timescale: str, dry_run=False) -> None:
        ...



from typing import Dict

from nimbo.core.cloud_provider.provider.services.instance import Instance


class GcpInstance(Instance):
    @staticmethod
    def run(job_cmd: str, dry_run=False) -> Dict[str, str]:
        pass

    @staticmethod
    def run_access_test(dry_run=False) -> None:
        pass

    @staticmethod
    def _block_until_instance_running(instance_id: str) -> None:
        pass

    @staticmethod
    def _get_host_from_instance_id(instance_id: str, dry_run=False) -> str:
        pass

    @staticmethod
    def stop_instance(instance_id: str, dry_run=False) -> None:
        pass

    @staticmethod
    def resume_instance(instance_id: str, dry_run=False) -> None:
        pass

    @staticmethod
    def delete_instance(instance_id: str, dry_run=False) -> None:
        pass

    @staticmethod
    def delete_all_instances(dry_run=False) -> None:
        pass

    @staticmethod
    def get_status(instance_id: str, dry_run=False) -> str:
        pass

    @staticmethod
    def ls_active_instances(dry_run=False) -> None:
        pass

    @staticmethod
    def ls_stopped_instances(dry_run=False) -> None:
        pass



import os

import click

from nimbo.core.constants import NIMBO_DEFAULT_CONFIG


def generate_config(config_path: str) -> None:
    """ Create an example Nimbo config in the project root """

    if os.path.isfile(config_path):
        should_overwrite = click.confirm(
            f"{config_path} already exists, do you want to overwrite it?"
        )
        if not should_overwrite:
            print("Leaving Nimbo config intact")
            return

    with open(config_path, "w") as f:
        f.write(NIMBO_DEFAULT_CONFIG)

    print(f"Example config written to {config_path}")



import enum
import functools
import os
import shutil

from click.testing import CliRunner

from nimbo import CONFIG
from nimbo.core.config import RequiredCase
from nimbo.tests.aws.config import ASSETS_PATH, CONDA_ENV


def make_file(path: str, text: str) -> None:
    with open(path, "w") as f:
        f.write(text)


class AssetType(enum.Enum):
    NIMBO_CONFIG = 0
    INSTANCE_KEYS = 1
    CONDA_ENV = 2


def isolated_filesystem(*cases: RequiredCase):
    """
    Decorator for creating an isolated filesystem,
    copying required files and injecting required config based on RequiredCases

    Requires dummy fixture
    @pytest.fixture
    def runner():
        return ""

    to be specified in each file where this decorator is used.

    @isolated_filesystem passes the click CliRunner as an argument to the function
    that it wraps. Pytest expects all parameters to test functions be fixtures.
    This dummy fixture fools pytest, but in runtime, CliRunner is injected by
    @isolated_filesystem.
    """

    def decorator(func):
        # noinspection PyUnusedLocal
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            runner = CliRunner()
            with runner.isolated_filesystem():
                _setup_for_case(*cases)
                return func(runner)

        return decorated

    return decorator


def _setup_for_case(*cases: RequiredCase) -> None:
    """
    Used within CliRunner().isolated_filesystem() for copying what is needed
    from the test assets folder to the filesystem and for injecting the
    testing configuration needed for this particular case
    """

    cases = RequiredCase.decompose(*cases)

    if RequiredCase.MINIMAL in cases:
        _copy_assets(AssetType.NIMBO_CONFIG)
    if RequiredCase.INSTANCE in cases:
        _copy_assets(AssetType.INSTANCE_KEYS)
    if RequiredCase.JOB in cases:
        _copy_assets(AssetType.CONDA_ENV)

    CONFIG.reset_required_config()
    CONFIG.inject_required_config(*cases)


def _copy_assets(*assets: AssetType) -> None:
    dst = os.getcwd()

    if AssetType.NIMBO_CONFIG in assets:
        src = os.path.join(ASSETS_PATH, CONFIG.config_path)
        shutil.copy(src, dst)
    if AssetType.INSTANCE_KEYS in assets:
        keys = [file for file in os.listdir(ASSETS_PATH) if file[-4:] == ".pem"]
        for key in keys:
            shutil.copy(os.path.join(ASSETS_PATH, key), dst)
    if AssetType.CONDA_ENV in assets:
        shutil.copy(os.path.join(ASSETS_PATH, CONDA_ENV), dst)



from nimbo.core.cloud_provider.provider.services.storage import Storage


class GcpStorage(Storage):
    @staticmethod
    def push(folder: str, delete=False) -> None:
        ...

    @staticmethod
    def pull(folder: str, delete=False) -> None:
        ...

    @staticmethod
    def ls_bucket(path: str) -> None:
        ...

    @staticmethod
    def mk_bucket(bucket_name: str, dry_run=False) -> None:
        ...



import enum
import os
from typing import Optional, Set

import pydantic

from nimbo.core.constants import TELEMETRY_URL


class RequiredCase(str, enum.Enum):
    # First digit is a unique ID, other digits are the IDs of dependencies
    NONE = "0"
    MINIMAL = "10"
    STORAGE = "210"
    INSTANCE = "310"
    JOB = "43210"

    @classmethod
    def decompose(cls, *cases: "RequiredCase") -> Set["RequiredCase"]:
        """ Gets all cases that compose each case and the case itself """

        decomposed = set()

        for case in cases:
            for c in RequiredCase:
                if c[0] in case.value:
                    decomposed.add(c)

        # noinspection PyTypeChecker
        return decomposed


class CloudProvider(str, enum.Enum):
    AWS = "AWS"
    GCP = "GCP"


class BaseConfig(pydantic.BaseModel):
    class Config:
        title = "Nimbo configuration"
        extra = "forbid"

    config_path: str
    cloud_provider: CloudProvider = None

    local_datasets_path: Optional[str] = None
    local_results_path: Optional[str] = None

    conda_env: Optional[str] = None
    run_in_background: bool = False
    persist: bool = False

    ip_cidr_range: pydantic.conint(strict=True, ge=0, le=32) = 32
    ssh_timeout: pydantic.conint(strict=True, ge=0) = 180
    telemetry: bool = True

    user_id: Optional[str] = None
    telemetry_url: str = TELEMETRY_URL

    def _local_results_not_outside_project(self) -> Optional[str]:
        if os.path.isabs(self.local_results_path):
            return "local_results_path should be a relative path"
        if ".." in self.local_results_path:
            return "local_results_path should not be outside of the project directory"

    def _local_datasets_not_outside_project(self) -> Optional[str]:
        if os.path.isabs(self.local_datasets_path):
            return "local_datasets_path should be a relative path"
        if ".." in self.local_datasets_path:
            return "local_datasets_path should not be outside of the project directory"

    @pydantic.validator("telemetry_url")
    def _nimbo_telemetry_url_unchanged(cls, value):
        if value != TELEMETRY_URL:
            raise ValueError("overriding telemetry url is forbidden")
        return value



import abc


class Permissions(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def mk_instance_key(dry_run=False) -> None:
        """Create and download an instance key to the current directory."""
        ...

    @staticmethod
    @abc.abstractmethod
    def allow_ingress_current_ip(target: str, dry_run=False) -> None:
        """
        Adds the IP of the current machine to the allowed ingress rules of
        instances that

        :param target: group name for AWS or VPC name for GCP
        :param dry_run: perform dry run
        """
        ...

    @staticmethod
    @abc.abstractmethod
    def setup(profile: str, no_s3_access=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def add_user(profile: str, username: str) -> None:
        ...



import enum
import functools
import sys

import botocore.errorfactory
import click

from nimbo.core.constants import IS_TEST_ENV
from nimbo.core.print import nprint


class HelpSection(enum.Enum):
    INSTANCE = "Instance"
    STORAGE = "Storage"
    UTILS = "Utilities"
    ADMIN = "Admin Utilities"


class NimboCommand(click.Command):
    """ click.Command extension for command grouping """

    def __init__(self, *args, **kwargs):
        self.help_section: HelpSection = kwargs.pop("help_section", None)
        super().__init__(*args, **kwargs)


class NimboGroup(click.Group):
    """ click.Group extension for command grouping """

    def list_commands(self, ctx):
        # Avoid command sorting
        return self.commands.keys()

    def format_commands(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        # Display commands in sections

        section_to_commands = {section.value: [] for section in HelpSection}

        max_cmd_len = 0
        for sub_cmd in self.list_commands(ctx):
            max_cmd_len = max(max_cmd_len, len(sub_cmd))
            cmd = self.get_command(ctx, sub_cmd)

            if hasattr(cmd, "help_section") and cmd.help_section:
                section_to_commands[cmd.help_section.value].append((sub_cmd, cmd))

        # allow for 3 times the default spacing
        limit = formatter.width - 6 - max_cmd_len

        for section, commands in section_to_commands.items():
            if len(commands):
                rows = []
                max_section_cmd_len = 0

                for sub_cmd, cmd in commands:
                    max_section_cmd_len = max(max_section_cmd_len, len(sub_cmd))

                    help_text = cmd.get_short_help_str(limit)
                    rows.append((sub_cmd, help_text))

                if rows:
                    col_spacing = max_cmd_len - max_section_cmd_len + 2

                    with formatter.section(section):
                        formatter.write_dl(rows, col_spacing=col_spacing)


def pprint_errors(func):
    """
    Decorator for catching boto3 ClientErrors, ValueError or KeyboardInterrupts.
    In case of error print the error message and stop Nimbo.
    """

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        if IS_TEST_ENV:
            return func(*args, **kwargs)
        else:
            try:
                return func(*args, **kwargs)
            except botocore.errorfactory.ClientError as e:
                nprint(e, style="error")
                sys.exit(1)
            except ValueError as e:
                nprint(e, style="error")
                sys.exit(1)
            except KeyboardInterrupt:
                print("Aborting...")
                sys.exit(1)

    return decorated



import os
import re
from typing import Any, Dict

import pydantic
import yaml

# pattern for extracting env variables
from nimbo.core.config.common_config import BaseConfig

RE_PATTERN = re.compile(
    r"(\$(?:{(?P<env>(.*?))(\|(?P<env_default>.*?))?}))",
    re.MULTILINE | re.UNICODE | re.IGNORECASE | re.VERBOSE,
)


def _substitute_env_vars(key: str, value: str) -> str:
    """
    Take in a string optionally containing "${ENV_VARIABLE|optional-default}"
    and return a new string with substituted environment variables
    """

    replacements = []

    for env_var in RE_PATTERN.finditer(value):
        groups = env_var.groupdict()
        variable, default = groups["env"], groups["env_default"]

        to_replace = "${" + variable
        to_replace += "|" + default if default else ""
        to_replace += "}"

        replace_with = default
        try:
            replace_with = os.environ[variable]
        except KeyError:
            if not replace_with:
                raise pydantic.ValidationError(
                    [
                        pydantic.error_wrappers.ErrorWrapper(
                            Exception(f"Environment variable {variable} not defined"),
                            key,
                        )
                    ],
                    BaseConfig,
                )

        replacements.append([to_replace, replace_with])

    if replacements:
        for to_replace, replace_with in replacements:
            value = value.replace(to_replace, replace_with)

    return value


def from_file(file: str) -> Dict[str, Any]:
    """ Load YAML into a dictionary, inject environment variables """

    if os.path.isfile(file):
        with open(file, "r") as f:
            config = yaml.safe_load(f)

        for key, value in config.items():
            if type(value) == str and "${" in value:
                config[key] = _substitute_env_vars(key, value)

        return config

    return {}



import subprocess
import sys
import time
from pathlib import Path
from pprint import pprint
from typing import Dict, List, Union

import botocore.exceptions
import requests

from nimbo import CONFIG
from nimbo.core import telemetry
from nimbo.core.cloud_provider.provider.services.instance import Instance
from nimbo.core.cloud_provider.provider_impl.aws.services.aws_permissions import (
    AwsPermissions,
)
from nimbo.core.cloud_provider.provider_impl.aws.services.aws_storage import AwsStorage
from nimbo.core.constants import NIMBO_VARS
from nimbo.core.print import nprint, nprint_header


class AwsInstance(Instance):
    @staticmethod
    def run(job_cmd: str, dry_run=False) -> Dict[str, str]:
        if dry_run:
            return {"message": job_cmd + "_dry_run"}

        # Launch instance with new volume for anaconda
        telemetry.record_event("run")

        start_t = time.monotonic()

        instance_id = AwsInstance._start_instance()

        try:
            # Wait for the instance to be running
            AwsInstance._block_until_instance_running(instance_id)
            end_t = time.monotonic()
            nprint_header(f"Instance running. ({round((end_t - start_t), 2)} s)")
            nprint_header(f"InstanceId: [green]{instance_id}[/green]")
            print()

            time.sleep(5)
            host = AwsInstance._get_host_from_instance_id(instance_id)

            AwsInstance._block_until_ssh_ready(host)

            if job_cmd == "_nimbo_launch":
                nprint_header(
                    f"Run [cyan]nimbo ssh {instance_id}[/cyan] to log onto the instance"
                )
                return {"message": job_cmd + "_success", "instance_id": instance_id}

            ssh = (
                f"ssh -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no'"
                " -o ServerAliveInterval=5 "
            )
            scp = f"scp -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no'"

            local_env = "/tmp/local_env.yml"
            user_conda_yml = CONFIG.conda_env
            # TODO: Replace this with shutil
            subprocess.check_output(f"cp {user_conda_yml} {local_env}", shell=True)

            # Send conda env yaml and setup scripts to instance
            print()
            nprint_header(f"Syncing conda, config, and setup files...")
            AwsInstance._write_nimbo_vars()

            # Create project folder and send env and config files there
            subprocess.check_output(f"{ssh} ubuntu@{host} mkdir project", shell=True)
            subprocess.check_output(
                f"{scp} {local_env} {CONFIG.config_path} {NIMBO_VARS}"
                f" ubuntu@{host}:/home/ubuntu/project/",
                shell=True,
            )

            # Sync code with instance
            print()
            nprint_header(f"Syncing code...")
            AwsInstance._sync_code(host)

            nprint_header(f"Running setup code on the instance from here on.")
            # Run remote_setup script on instance
            AwsInstance._run_remote_script(
                ssh, scp, host, instance_id, job_cmd, "remote_setup.sh"
            )

            if job_cmd == "_nimbo_notebook":
                subprocess.Popen(
                    f"{ssh} -o 'ExitOnForwardFailure yes' "
                    f"ubuntu@{host} -NfL 57467:localhost:57467 >/dev/null 2>&1",
                    shell=True,
                ).communicate()
                nprint_header(
                    "Make sure to run 'nimbo sync-notebooks <instance_id>' frequently "
                    "to sync the notebook to your local folder, as the remote notebooks"
                    " will be lost once the instance is terminated."
                )

            return {"message": job_cmd + "_success", "instance_id": instance_id}

        except BaseException as e:
            if (
                type(e) != KeyboardInterrupt
                and type(e) != subprocess.CalledProcessError
            ):
                nprint(e, style="error")

            if not CONFIG.persist:
                nprint_header(f"Deleting instance {instance_id} (from local)... ")
                AwsInstance.delete_instance(instance_id)

            return {"message": job_cmd + "_error", "instance_id": instance_id}

    @staticmethod
    def run_access_test(dry_run=False) -> None:
        if dry_run:
            return

        CONFIG.instance_type = "t3.medium"
        CONFIG.run_in_background = False
        CONFIG.persist = False

        try:
            # Send test file to s3 results path and delete it
            profile = CONFIG.aws_profile
            region = CONFIG.region_name
            results_path = CONFIG.s3_results_path

            subprocess.check_output(
                "echo 'Hello World' > nimbo-access-test.txt", shell=True
            )
            command = AwsStorage.mk_s3_command(
                "cp", "nimbo-access-test.txt", results_path
            )
            subprocess.check_output(command, shell=True)

            command = f"aws s3 ls {results_path} --profile {profile} --region {region}"
            subprocess.check_output(command, shell=True)
            command = (
                f"aws s3 rm {results_path}/nimbo-access-test.txt "
                f"--profile {profile} --region {region}"
            )
            subprocess.check_output(command, shell=True)

            print(
                "You have the necessary S3 read/write "
                "permissions from your computer \u2713"
            )

        except subprocess.CalledProcessError as e:
            nprint(e, style="error")
            sys.exit(1)

        # Launch instance with new volume for anaconda
        print("Launching test instance... ")

        instance_id = AwsInstance._start_instance()

        try:
            # Wait for the instance to be running
            AwsInstance._block_until_instance_running(instance_id)
            print(f"Instance running. Instance creation allowed \u2713")
            print(f"InstanceId: {instance_id}")
            print()

            print("Trying to delete this instance...")
            AwsInstance.delete_instance(instance_id)

            print("Instance deletion allowed \u2713")
            print("\nLaunching another instance...")
            instance_id = AwsInstance._start_instance()
            print(f"Instance running. InstanceId: {instance_id}")

            time.sleep(5)
            host = AwsInstance._get_host_from_instance_id(instance_id)
            ssh = (
                f"ssh -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no' "
                "-o ServerAliveInterval=20"
            )
            scp = f"scp -i {CONFIG.instance_key} -o 'StrictHostKeyChecking no'"

            AwsInstance._block_until_ssh_ready(host)

            print("Instance key allows ssh access to remote instance \u2713")
            print("Security group allows ssh access to remote instance \u2713")

            AwsInstance._write_nimbo_vars()

            subprocess.check_output(
                f"{scp} {CONFIG.config_path} {NIMBO_VARS} "
                + f"ubuntu@{host}:/home/ubuntu/",
                shell=True,
            )
            AwsInstance._run_remote_script(
                ssh, scp, host, instance_id, "", "remote_s3_test.sh"
            )

        except BaseException as e:
            if (
                type(e) != KeyboardInterrupt
                and type(e) != subprocess.CalledProcessError
            ):
                nprint(e, style="error")

            if not CONFIG.persist:
                nprint_header(f"Deleting instance {instance_id} (from local)...")
                AwsInstance.delete_instance(instance_id)

            sys.exit(1)

    @staticmethod
    def _block_until_instance_running(instance_id: str) -> None:
        status = ""
        while status != "running":
            time.sleep(1)
            status = AwsInstance.get_status(instance_id)

    @staticmethod
    def _write_nimbo_vars() -> None:
        var_list = [
            f"S3_DATASETS_PATH={CONFIG.s3_datasets_path}",
            f"S3_RESULTS_PATH={CONFIG.s3_results_path}",
            f"LOCAL_DATASETS_PATH={CONFIG.local_datasets_path}",
            f"LOCAL_RESULTS_PATH={CONFIG.local_results_path}",
        ]
        if CONFIG.encryption:
            var_list.append(f"ENCRYPTION={CONFIG.encryption}")
        with open(NIMBO_VARS, "w") as f:
            f.write("\n".join(var_list))

    @staticmethod
    def _get_host_from_instance_id(instance_id: str, dry_run=False) -> str:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.describe_instances(
                InstanceIds=[instance_id],
                Filters=AwsInstance._make_instance_filters(),
                DryRun=dry_run,
            )
            host = response["Reservations"][0]["Instances"][0]["PublicIpAddress"]
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise
            host = ""
        return host

    @staticmethod
    def _get_image_id() -> str:
        if CONFIG.image[:4] == "ami-":
            image_id = CONFIG.image
        else:
            response = requests.get(
                "https://nimboami-default-rtdb.firebaseio.com/images.json"
            )
            catalog = response.json()
            region = CONFIG.region_name
            if region in catalog:
                region_catalog = catalog[region]
                image_name = CONFIG.image
                if image_name in region_catalog:
                    image_id = region_catalog[image_name]
                else:
                    raise ValueError(
                        f"The image {image_name} was not found in the"
                        " image catalog managed by Nimbo.\n"
                        "Check https://docs.nimbo.sh/managed-images"
                        " for the list of managed images."
                    )
            else:
                raise ValueError(
                    f"Currently, Nimbo does not support managed images in {region}."
                    " Please use another region."
                )

        return image_id

    @staticmethod
    def _make_instance_tags() -> List[Dict[str, str]]:
        return [
            {"Key": "CreatedBy", "Value": "nimbo"},
            {"Key": "Owner", "Value": CONFIG.user_id},
        ]

    @staticmethod
    def _make_instance_filters() -> List[Dict[str, Union[str, List[str]]]]:
        tags = AwsInstance._make_instance_tags()
        filters = []
        for tag in tags:
            tag_filter = {"Name": "tag:" + tag["Key"], "Values": [tag["Value"]]}
            filters.append(tag_filter)
        return filters

    @staticmethod
    def _start_instance() -> str:
        AwsPermissions.allow_ingress_current_ip(CONFIG.security_group)

        ec2 = CONFIG.get_session().client("ec2")
        instance_tags = AwsInstance._make_instance_tags()
        instance_filters = AwsInstance._make_instance_filters()

        image = AwsInstance._get_image_id()
        nprint_header(f"Launching instance with image {image}... ")

        ebs_config = {
            "VolumeSize": CONFIG.disk_size,
            "VolumeType": CONFIG.disk_type,
        }
        if CONFIG.disk_iops:
            ebs_config["Iops"] = CONFIG.disk_iops

        instance_config = {
            "BlockDeviceMappings": [{"DeviceName": "/dev/sda1", "Ebs": ebs_config}],
            "ImageId": image,
            "InstanceType": CONFIG.instance_type,
            "KeyName": Path(CONFIG.instance_key).stem,
            "Placement": {"Tenancy": "default"},
            "SecurityGroups": [CONFIG.security_group],
            "IamInstanceProfile": {"Name": CONFIG.role},
        }

        if CONFIG.spot:
            extra_kwargs = {}
            if CONFIG.spot_duration:
                extra_kwargs = {"BlockDurationMinutes": CONFIG.spot_duration}

            instance = ec2.request_spot_instances(
                LaunchSpecification=instance_config,
                TagSpecifications=[
                    {"ResourceType": "spot-instances-request", "Tags": instance_tags}
                ],
                **extra_kwargs,
            )
            instance_request = instance["SpotInstanceRequests"][0]
            request_id = instance_request["SpotInstanceRequestId"]

            try:
                nprint_header("Spot instance request submitted.")
                nprint_header(
                    "Waiting for the spot instance request to be fulfilled... "
                )

                status = ""
                while status != "fulfilled":
                    time.sleep(2)
                    response = ec2.describe_spot_instance_requests(
                        SpotInstanceRequestIds=[request_id],
                        Filters=instance_filters,
                    )
                    instance_request = response["SpotInstanceRequests"][0]
                    status = instance_request["Status"]["Code"]
                    if status not in [
                        "fulfilled",
                        "pending-evaluation",
                        "pending-fulfillment",
                    ]:
                        raise Exception(response["SpotInstanceRequests"][0]["Status"])
            except KeyboardInterrupt:
                ec2.cancel_spot_instance_requests(SpotInstanceRequestIds=[request_id])
                nprint_header("Cancelled spot instance request.")
                sys.exit(1)

            nprint_header("Done.")
            ec2.create_tags(
                Resources=[instance_request["InstanceId"]],
                Tags=instance_tags,
            )
            instance = instance_request
        else:
            instance_config["MinCount"] = 1
            instance_config["MaxCount"] = 1
            instance_config["InstanceInitiatedShutdownBehavior"] = "terminate"
            instance_config["TagSpecifications"] = [
                {"ResourceType": "instance", "Tags": instance_tags}
            ]
            instance = ec2.run_instances(**instance_config)
            instance = instance["Instances"][0]

        return instance["InstanceId"]

    @staticmethod
    def stop_instance(instance_id: str, dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.stop_instances(InstanceIds=[instance_id], DryRun=dry_run)
            pprint(response)
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise

    @staticmethod
    def resume_instance(instance_id: str, dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.start_instances(InstanceIds=[instance_id], DryRun=dry_run)
            pprint(response)
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise

    @staticmethod
    def delete_instance(instance_id: str, dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.terminate_instances(
                InstanceIds=[instance_id], DryRun=dry_run
            )
            status = response["TerminatingInstances"][0]["CurrentState"]["Name"]
            nprint_header(f"Instance [green]{instance_id}[/green]: {status}")
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise

    @staticmethod
    def delete_all_instances(dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.describe_instances(
                Filters=[{"Name": "instance-state-name", "Values": ["running"]}]
                + AwsInstance._make_instance_filters(),
                DryRun=dry_run,
            )
            for reservation in response["Reservations"]:
                for inst in reservation["Instances"]:
                    instance_id = inst["InstanceId"]
                    delete_response = ec2.terminate_instances(
                        InstanceIds=[instance_id],
                    )
                    status = delete_response["TerminatingInstances"][0]["CurrentState"][
                        "Name"
                    ]
                    nprint_header(f"Instance [green]{instance_id}[/green]: {status}")
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise

    @staticmethod
    def get_status(instance_id: str, dry_run=False) -> str:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.describe_instances(
                InstanceIds=[instance_id],
                Filters=AwsInstance._make_instance_filters(),
                DryRun=dry_run,
            )
            status = response["Reservations"][0]["Instances"][0]["State"]["Name"]
            return status
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise

    @staticmethod
    def ls_active_instances(dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.describe_instances(
                Filters=[
                    {"Name": "instance-state-name", "Values": ["running", "pending"]}
                ]
                + AwsInstance._make_instance_filters(),
                DryRun=dry_run,
            )
            for reservation in response["Reservations"]:
                for inst in reservation["Instances"]:
                    nprint(
                        f"Id: [bright_green]{inst['InstanceId']}[/bright_green]\n"
                        f"Status: {inst['State']['Name']}\n"
                        f"Launch Time: {inst['LaunchTime']}\n"
                        f"InstanceType: {inst['InstanceType']}\n"
                        f"IP Address: {inst['PublicIpAddress']}\n"
                    )

        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise

    @staticmethod
    def ls_stopped_instances(dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")
        try:
            response = ec2.describe_instances(
                Filters=[
                    {"Name": "instance-state-name", "Values": ["stopped", "stopping"]}
                ]
                + AwsInstance._make_instance_filters(),
                DryRun=dry_run,
            )
            for reservation in response["Reservations"]:
                for inst in reservation["Instances"]:
                    print(
                        f"ID: {inst['InstanceId']}\n"
                        f"Launch Time: {inst['LaunchTime']}\n"
                        f"InstanceType: {inst['InstanceType']}\n"
                    )
        except botocore.exceptions.ClientError as e:
            if "DryRunOperation" not in str(e):
                raise



import json
import os
import sys

import boto3
import botocore.exceptions
import requests

from nimbo import CONFIG
from nimbo.core.cloud_provider.provider.services.permissions import Permissions
from nimbo.core.constants import ASSUME_ROLE_POLICY, EC2_POLICY_JSON
from nimbo.core.print import nprint, nprint_header

NIMBO_USER_GROUP = "NimboUserGroup"
EC2_POLICY_NAME = "NimboEC2Policy"
CRED_POLICY_NAME = "NimboCredentialsPolicy"
PASS_ROLE_POLICY_NAME = "NimboPassRolePolicy"
S3_ACCESS_ROLE_NAME = "NimboFullS3AccessRole"


class AwsPermissions(Permissions):
    @staticmethod
    def mk_instance_key(dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")

        if "/" in CONFIG.user_arn:
            username = CONFIG.user_arn.split("/")[1]
        else:
            username = CONFIG.user_arn.split(":")[-1]
        key_name = f"{username}-{CONFIG.region_name}"

        try:
            response = ec2.create_key_pair(KeyName=key_name, DryRun=dry_run)
            key_body = response["KeyMaterial"]

            with open(f"{key_name}.pem", "w") as f:
                f.write(key_body)

            os.chmod(f"{key_name}.pem", int("400", base=8))
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "UnauthorizedOperation":
                return
            else:
                raise

    @staticmethod
    def allow_ingress_current_ip(target: str, dry_run=False) -> None:
        ec2 = CONFIG.get_session().client("ec2")

        try:
            response = ec2.describe_security_groups(GroupNames=[target], DryRun=dry_run)
            security_group_id = response["SecurityGroups"][0]["GroupId"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "InvalidGroup.NotFound":
                nprint(
                    f"Security group {target} not found. Please use an existing"
                    " security group or create a new one in the AWS console.",
                    style="error",
                )
                sys.exit(1)
            elif e.response["Error"]["Code"] == "UnauthorizedOperation":
                return
            else:
                raise

        my_public_ip = requests.get("https://checkip.amazonaws.com").text.strip()

        try:
            ec2.authorize_security_group_ingress(
                GroupId=security_group_id,
                IpPermissions=[
                    {
                        "IpProtocol": "tcp",
                        "FromPort": 22,
                        "ToPort": 22,
                        "IpRanges": [
                            {"CidrIp": f"{my_public_ip}/{CONFIG.ip_cidr_range}"}
                        ],
                    }
                ],
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "InvalidPermission.Duplicate":
                return
            elif e.response["Error"]["Code"] == "UnauthorizedOperation":
                return
            else:
                raise

    @staticmethod
    def setup(profile: str, no_s3_access=False) -> None:
        session = boto3.Session(profile_name=profile)
        account = session.client("sts").get_caller_identity()["Account"]

        iam = session.client("iam")

        nprint_header(f"Creating user group {NIMBO_USER_GROUP}...")
        AwsPermissions._create_group(iam, NIMBO_USER_GROUP)

        nprint_header(f"Creating policy {EC2_POLICY_NAME}...")
        AwsPermissions._create_policy(iam, EC2_POLICY_NAME, EC2_POLICY_JSON)

        nprint_header(
            f"Attaching policy {EC2_POLICY_NAME} to user group {NIMBO_USER_GROUP}..."
        )
        iam.attach_group_policy(
            GroupName=NIMBO_USER_GROUP,
            PolicyArn=f"arn:aws:iam::{account}:policy/{EC2_POLICY_NAME}",
        )

        if no_s3_access:
            nprint(
                "\nSince you chose not to give full S3 access to the Nimbo user group"
                " and instance role,\nwe recommend that you create a role with the"
                " necessary S3 permissions in the AWS console.\nOnce you do this, give"
                " the role name to the people using Nimbo so that they can set\n"
                "the 'role' field in the configuration file to this value.",
                style="warning",
            )
        else:
            nprint_header(f"Creating role {S3_ACCESS_ROLE_NAME}...")
            AwsPermissions._create_role_and_instance_profile(iam, S3_ACCESS_ROLE_NAME)

            nprint_header(
                f"Attaching AmazonS3FullAccess policy to role {S3_ACCESS_ROLE_NAME}..."
            )
            iam.attach_role_policy(
                PolicyArn="arn:aws:iam::aws:policy/AmazonS3FullAccess",
                RoleName=S3_ACCESS_ROLE_NAME,
            )

            nprint_header(f"Creating policy {PASS_ROLE_POLICY_NAME}...")
            pass_role_policy_json = {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "NimboPassRolePolicy",
                        "Effect": "Allow",
                        "Action": "iam:PassRole",
                        "Resource": f"arn:aws:iam::*:role/{S3_ACCESS_ROLE_NAME}",
                    }
                ],
            }
            AwsPermissions._create_policy(
                iam, PASS_ROLE_POLICY_NAME, pass_role_policy_json
            )

            nprint_header(
                f"Attaching policy {PASS_ROLE_POLICY_NAME}"
                f" to user group {NIMBO_USER_GROUP}..."
            )
            iam.attach_group_policy(
                GroupName=NIMBO_USER_GROUP,
                PolicyArn=f"arn:aws:iam::{account}:policy/{PASS_ROLE_POLICY_NAME}",
            )

            nprint_header(
                f"Attaching policy AmazonS3FullAccess"
                f" to user group {NIMBO_USER_GROUP}..."
            )
            iam.attach_group_policy(
                GroupName=NIMBO_USER_GROUP,
                PolicyArn="arn:aws:iam::aws:policy/AmazonS3FullAccess",
            )

        print()
        nprint_header("Done.")
        nprint_header(
            "To add users to the NimboUserGroup, simply"
            " run 'nimbo add-user USERNAME YOUR_AWS_PROFILE'.\n"
            "For more info use 'nimbo add-user --help'"
        )

    @staticmethod
    def add_user(profile: str, username: str) -> None:
        session = boto3.Session(profile_name=profile)
        iam = session.client("iam")

        iam.add_user_to_group(GroupName=NIMBO_USER_GROUP, UserName=username)
        print(f"User {username} added to {NIMBO_USER_GROUP}.")

    @staticmethod
    def _create_group(client, group_name):
        try:
            client.create_group(GroupName=group_name)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityAlreadyExists":
                nprint(
                    f"User group {group_name} already exists. Skipping.",
                    style="warning",
                )
            else:
                raise

    @staticmethod
    def _create_policy(client, policy_name, policy_json):
        try:
            client.create_policy(
                PolicyName=policy_name, PolicyDocument=json.dumps(policy_json)
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityAlreadyExists":
                nprint(
                    f"Policy {policy_name} already exists. Skipping.", style="warning"
                )
            else:
                raise

    @staticmethod
    def _create_role_and_instance_profile(client, role_name):
        try:
            client.create_role(
                RoleName=role_name,
                AssumeRolePolicyDocument=json.dumps(ASSUME_ROLE_POLICY),
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityAlreadyExists":
                nprint(f"Role {role_name} already exists. Skipping.", style="warning")
            else:
                raise

        try:
            client.create_instance_profile(InstanceProfileName=role_name, Path="/")
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityAlreadyExists":
                nprint(
                    f"Instance profile for role {role_name} already exists. Skipping.",
                    style="warning",
                )
            else:
                raise

        try:
            client.add_role_to_instance_profile(
                InstanceProfileName=role_name, RoleName=role_name
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "LimitExceeded":
                nprint(
                    f"Instance profile {role_name} already has a role. Skipping.",
                    style="warning",
                )
            else:
                raise



from nimbo.core.cloud_provider.provider_impl.gcp.services import (
    gcp_instance,
    gcp_permissions,
    gcp_storage,
    gcp_utils,
)


class GcpProvider(
    gcp_instance.GcpInstance,
    gcp_permissions.GcpPermissions,
    gcp_storage.GcpStorage,
    gcp_utils.GcpUtils,
):
    ...



import os
import pathlib

IS_TEST_ENV = "NIMBO_ENV" in os.environ and os.environ["NIMBO_ENV"] == "test"

NIMBO_ROOT = str(pathlib.Path(__file__).parent.parent.absolute())
NIMBO_VARS = "/tmp/nimbo_vars"
TELEMETRY_URL = "https://nimbotelemetry-8ef4c-default-rtdb.firebaseio.com/events.json"

NIMBO_DEFAULT_CONFIG = """cloud_provider: AWS

# Data paths
local_datasets_path: your-datasets-folder  # relative to project root
local_results_path: your-results-folder    # relative to project root
s3_datasets_path: s3://your-bucket/your-project/example-datasets-folder
s3_results_path: s3://your-bucket/your-project/example-results-folder

# Device, environment and regions
aws_profile: default
region_name: eu-west-1
instance_type: p2.xlarge
spot: no

image: ubuntu18-latest-drivers
disk_size: 128  # In GB
conda_env: your-env-file.yml  # denotes project root

# Job options
run_in_background: no
persist: no  # whether instance persists when the job finishes or on error

# Permissions and credentials
security_group: default
instance_key: your-ec2-key-pair.pem  # can be an absolute path
role: NimboFullS3AccessRole
"""


# Each element is [num_gpus, gpu_type, ram, vcpus]
INSTANCE_GPU_MAP = {
    "p4d.24xlarge": [8, "A100", 1152, 96],
    "p3.2xlarge": [1, "V100", 61, 8],
    "p3.8xlarge": [4, "V100", 244, 32],
    "p3.16xlarge": [8, "V100", 488, 64],
    "p3dn.24xlarge": [8, "V100", 768, 96],
    "p2.xlarge": [1, "K80", 61, 4],
    "p2.8xlarge": [8, "K80", 488, 32],
    "p2.16xlarge": [16, "K80", 732, 64],
    "g4dn.xlarge": [1, "T4", 16, 4],
    "g4dn.2xlarge": [1, "T4", 32, 8],
    "g4dn.4xlarge": [1, "T4", 64, 16],
    "g4dn.8xlarge": [1, "T4", 128, 32],
    "g4dn.16xlarge": [1, "T4", 256, 64],
    "g4dn.12xlarge": [4, "T4", 192, 48],
    "g4dn.metal": [8, "T4", 384, 96],
}


FULL_REGION_NAMES = {
    "af-south-1": "Africa (Cape Town)",
    "ap-east-1": "Asia Pacific (Hong Kong)",
    "ap-south-1": "Asia Pacific (Mumbai)",
    "ap-northeast-3": "Asia Pacific (Osaka)",
    "ap-northeast-2": "Asia Pacific (Seoul)",
    "ap-southeast-1": "Asia Pacific (Singapore)",
    "ap-southeast-2": "Asia Pacific (Sydney)",
    "ap-northeast-1": "Asia Pacific (Tokyo)",
    "ca-central-1": "Canada (Central)",
    "eu-central-1": "EU (Frankfurt)",
    "eu-west-1": "EU (Ireland)",
    "eu-west-2": "EU (London)",
    "eu-south-1": "EU (Milan)",
    "eu-west-3": "EU (Paris)",
    "eu-north-1": "EU (Stockholm)",
    "me-south-1": "Middle East (Bahrain)",
    "sa-east-1": "South America (Sao Paulo)",
    "us-east-1": "US East (N. Virginia)",
    "us-east-2": "US East (Ohio)",
    "us-west-1": "US West (N. California)",
    "us-west-2": "US West (Oregon)",
}


ASSUME_ROLE_POLICY = {
    "Version": "2012-10-17",
    "Statement": {
        "Effect": "Allow",
        "Action": "sts:AssumeRole",
        "Principal": {"Service": "ec2.amazonaws.com"},
    },
}


EC2_POLICY_JSON = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "NimboEC2Policy1",
            "Effect": "Allow",
            "Action": [
                "ec2:AuthorizeSecurityGroupEgress",
                "ec2:AuthorizeSecurityGroupIngress",
                "ec2:CreateKeyPair",
                "ec2:CreateSecurityGroup",
                "ec2:CreateTags",
                "ec2:CreateVolume",
                "ec2:DeleteKeyPair",
                "ec2:DeleteSecurityGroup",
                "ec2:DeleteTags",
                "ec2:DeleteVolume",
                "ec2:Describe*",
                "ec2:GetConsole*",
                "ec2:ModifySnapshotAttribute",
                "ec2:RequestSpotInstances",
                "ec2:RevokeSecurityGroupEgress",
                "ec2:RevokeSecurityGroupIngress",
                "ec2:RunInstances",
            ],
            "Resource": "*",
        },
        {
            "Sid": "NimboEC2Policy2",
            "Effect": "Allow",
            "Action": [
                "ec2:StartInstances",
                "ec2:StopInstances",
                "ec2:TerminateInstances",
                "ec2:CancelSpotInstanceRequests",
            ],
            "Resource": "*",
            "Condition": {"StringEquals": {"ec2:ResourceTag/Owner": "${aws:userid}"}},
        },
        {
            "Sid": "NimboEC2Policy3",
            "Effect": "Allow",
            "Action": ["ec2:AttachVolume", "ec2:DetachVolume"],
            "Resource": "arn:aws:ec2:*:*:instance/*",
            "Condition": {"StringEquals": {"ec2:ResourceTag/Owner": "${aws:userid}"}},
        },
        {
            "Sid": "NimboPricingPolicy",
            "Effect": "Allow",
            "Action": ["pricing:*"],
            "Resource": "*",
        },
    ],
}



import os
from os.path import join

import pytest
from botocore.exceptions import ClientError
from click.testing import CliRunner

from nimbo import CONFIG
from nimbo.core.config import RequiredCase
from nimbo.main import cli
from nimbo.tests.aws.utils import isolated_filesystem, make_file


@pytest.fixture
def runner():
    return ""


@isolated_filesystem(RequiredCase.NONE)
def test_generate_config(runner: CliRunner):
    result = runner.invoke(cli, "generate-config", catch_exceptions=False)
    assert result.exit_code == 0, result.exception


@isolated_filesystem(RequiredCase.MINIMAL)
def test_ssh_fails_with_minimal(runner: CliRunner):
    result = runner.invoke(
        cli, "ssh i-0be1989edd819b442 --dry-run", catch_exceptions=False
    )
    assert result.exit_code == 1


@isolated_filesystem(RequiredCase.INSTANCE)
def test_ssh_passes_with_instance(runner: CliRunner):
    result = runner.invoke(
        cli, "ssh i-0be1989edd819b442 --dry-run", catch_exceptions=False
    )
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.MINIMAL)
def test_list_prices(runner: CliRunner):
    result = runner.invoke(cli, "ls-prices", catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(cli, "ls-spot-prices", catch_exceptions=False)
    assert result.exit_code == 0

    # Check if it works for us-east-2 region
    CONFIG.region_name = "us-east-2"

    result = runner.invoke(cli, "ls-prices", catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(cli, "ls-spot-prices", catch_exceptions=False)
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.MINIMAL)
def test_list_instances(runner: CliRunner):
    result = runner.invoke(cli, "ls-active --dry-run", catch_exceptions=False)
    assert result.exit_code == 0

    result = runner.invoke(cli, "ls-stopped --dry-run", catch_exceptions=False)
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.MINIMAL)
def test_instance_actions(runner: CliRunner):
    try:
        runner.invoke(
            cli,
            "get-status i-0be1989edd819b442 --dry-run",
            catch_exceptions=False,
        )
    except ClientError as e:
        if "InvalidInstanceID.NotFound" not in str(e):
            raise

    try:
        runner.invoke(
            cli,
            "stop-instance i-0be1989edd819b442 --dry-run",
            catch_exceptions=False,
        )
    except ClientError as e:
        if "InvalidInstanceID.NotFound" not in str(e):
            raise

    try:
        runner.invoke(
            cli,
            "rm-instance i-0be1989edd819b442 --dry-run",
            catch_exceptions=False,
        )
    except ClientError as e:
        if "InvalidInstanceID.NotFound" not in str(e):
            raise

    result = runner.invoke(
        cli, "rm-all-instances --dry-run", input="y", catch_exceptions=False
    )
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.JOB)
def test_run_job(runner: CliRunner):
    result = runner.invoke(
        cli, "run 'python --version' --dry-run", catch_exceptions=False
    )
    assert result.exit_code == 0

    result = runner.invoke(cli, "launch --dry-run", catch_exceptions=False)
    assert result.exit_code == 0

    result = runner.invoke(cli, "launch-and-setup --dry-run", catch_exceptions=False)
    assert result.exit_code == 0

    result = runner.invoke(cli, "test-access --dry-run", catch_exceptions=False)
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.STORAGE)
def test_push_pull(runner: CliRunner):
    os.mkdir(CONFIG.local_datasets_path)
    os.mkdir(CONFIG.local_results_path)
    result = runner.invoke(
        cli, "push datasets --delete", input="y", catch_exceptions=False
    )
    assert result.exit_code == 0
    result = runner.invoke(
        cli, "push results --delete", input="y", catch_exceptions=False
    )
    assert result.exit_code == 0

    # Run the code below for both datasets and results folders
    for mode in ["datasets", "results"]:
        if mode == "datasets":
            folder = CONFIG.local_datasets_path
        else:
            folder = CONFIG.local_results_path

        file_name = join(folder, "mnist.txt")

        # Add a dataset to local and push it to S3
        make_file(file_name, "Mock data")
        result = runner.invoke(cli, f"push {mode}", catch_exceptions=False)
        assert result.exit_code == 0

        # Delete that datasets locally and import it from S3
        os.remove(file_name)
        result = runner.invoke(cli, f"pull {mode}", catch_exceptions=False)
        assert result.exit_code == 0
        assert os.listdir(folder) == ["mnist.txt"]

        # Delete that dataset locally and push --delete to S3
        os.remove(file_name)
        result = runner.invoke(
            cli, f"push {mode} --delete", input="y", catch_exceptions=False
        )
        assert result.exit_code == 0

        # Pull from S3 and check that the dataset is still deleted
        result = runner.invoke(cli, f"pull {mode}", catch_exceptions=False)
        assert result.exit_code == 0
        assert os.listdir(folder) == []

    logs_folder = join(CONFIG.local_results_path, "nimbo-logs")
    os.mkdir(logs_folder)
    file_name = join(logs_folder, "log.txt")
    make_file(file_name, "Mock log")
    assert os.listdir(logs_folder) == ["log.txt"]

    result = runner.invoke(cli, "push logs", catch_exceptions=False)
    assert result.exit_code == 0

    os.remove(file_name)
    assert os.listdir(logs_folder) == []

    result = runner.invoke(cli, "pull logs", catch_exceptions=False)
    assert result.exit_code == 0
    assert os.listdir(logs_folder) == ["log.txt"]

    os.remove(file_name)
    result = runner.invoke(cli, "push logs --delete", input="y", catch_exceptions=False)
    assert result.exit_code == 0
    os.rmdir(logs_folder)
    result = runner.invoke(
        cli, "push results --delete", input="y", catch_exceptions=False
    )
    assert result.exit_code == 0



from nimbo.core.cloud_provider.provider_impl.aws.services import (
    aws_instance,
    aws_permissions,
    aws_storage,
    aws_utils,
)


class AwsProvider(
    aws_instance.AwsInstance,
    aws_permissions.AwsPermissions,
    aws_storage.AwsStorage,
    aws_utils.AwsUtils,
):
    ...



import os.path
import subprocess

import botocore.exceptions

from nimbo import CONFIG
from nimbo.core.cloud_provider.provider.services.storage import Storage
from nimbo.core.print import nprint


class AwsStorage(Storage):
    # noinspection DuplicatedCode
    @staticmethod
    def push(folder: str, delete=False) -> None:
        assert folder in ["datasets", "results", "logs"]

        if folder == "logs":
            source = os.path.join(CONFIG.local_results_path, "nimbo-logs")
            target = os.path.join(CONFIG.s3_results_path, "nimbo-logs")
        else:
            if folder == "results":
                source = CONFIG.local_results_path
                target = CONFIG.s3_results_path
            else:
                source = CONFIG.local_datasets_path
                target = CONFIG.s3_datasets_path

        AwsStorage._sync_folder(source, target, delete)

    # noinspection DuplicatedCode
    @staticmethod
    def pull(folder: str, delete=False) -> None:
        assert folder in ["datasets", "results", "logs"]

        if folder == "logs":
            source = os.path.join(CONFIG.s3_results_path, "nimbo-logs")
            target = os.path.join(CONFIG.local_results_path, "nimbo-logs")
        else:
            if folder == "results":
                source = CONFIG.s3_results_path
                target = CONFIG.local_results_path
            else:
                source = CONFIG.s3_datasets_path
                target = CONFIG.local_datasets_path

        AwsStorage._sync_folder(source, target, delete)

    @staticmethod
    def ls_bucket(path: str) -> None:
        profile = CONFIG.aws_profile
        region = CONFIG.region_name
        path = path.rstrip("/") + "/"
        command = f"aws s3 ls {path} --profile {profile} --region {region}"
        print(f"Running command: {command}")
        subprocess.Popen(command, shell=True).communicate()

    @staticmethod
    def mk_bucket(bucket_name: str, dry_run=False) -> None:
        """Create an S3 bucket in a specified region

        :param bucket_name: Bucket to create
        :param dry_run
        :return: True if bucket created, else False
        """

        try:
            session = CONFIG.get_session()
            s3 = session.client("s3")
            location = {"LocationConstraint": session.region_name}
            s3.create_bucket(Bucket=bucket_name, CreateBucketConfiguration=location)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "BucketAlreadyOwnedByYou":
                nprint("Bucket nimbo-main-bucket already exists.", style="warning")
            else:
                nprint(e, style="error")
            return

        print("Bucket %s created." % bucket_name)

    @staticmethod
    def _sync_folder(source, target, delete=False) -> None:
        command = AwsStorage.mk_s3_command("sync", source, target, delete)
        print(f"\nRunning command: {command}")
        subprocess.Popen(command, shell=True).communicate()

    @staticmethod
    def mk_s3_command(cmd, source, target, delete=False) -> str:
        command = (
            f"aws s3 {cmd} {source} {target} "
            f" --profile {CONFIG.aws_profile} --region {CONFIG.region_name}"
        )

        if delete:
            command += " --delete"

        if CONFIG.encryption:
            command += f" --sse {CONFIG.encryption}"
        return command



import json
from datetime import date, datetime
from typing import Generator

from dateutil.relativedelta import relativedelta

from nimbo import CONFIG
from nimbo.core.cloud_provider.provider.services.utils import Utils
from nimbo.core.constants import FULL_REGION_NAMES, INSTANCE_GPU_MAP
from nimbo.core.print import nprint


class AwsUtils(Utils):
    @staticmethod
    def ls_gpu_prices(dry_run=False) -> None:
        if dry_run:
            return

        full_region_name = FULL_REGION_NAMES[CONFIG.region_name]

        pricing = CONFIG.get_session().client("pricing", region_name="us-east-1")

        string = AwsUtils._format_price_string(
            "InstanceType", "Price ($/hour)", "GPUs", "CPUs", "Mem (Gb)"
        )
        print()
        nprint(string, style="bold")

        for instance_type in AwsUtils._instance_types():
            response = pricing.get_products(
                ServiceCode="AmazonEC2",
                MaxResults=100,
                FormatVersion="aws_v1",
                Filters=[
                    {
                        "Type": "TERM_MATCH",
                        "Field": "instanceType",
                        "Value": instance_type,
                    },
                    {
                        "Type": "TERM_MATCH",
                        "Field": "location",
                        "Value": full_region_name,
                    },
                    {
                        "Type": "TERM_MATCH",
                        "Field": "operatingSystem",
                        "Value": "Linux",
                    },
                    {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
                    {"Type": "TERM_MATCH", "Field": "preInstalledSw", "Value": "NA"},
                    {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "shared"},
                ],
            )

            inst = json.loads(response["PriceList"][0])
            inst = inst["terms"]["OnDemand"]
            inst = list(inst.values())[0]
            inst = list(inst["priceDimensions"].values())[0]
            inst = inst["pricePerUnit"]
            currency = list(inst.keys())[0]
            price = float(inst[currency])

            num_gpus, gpu_type, mem, cpus = INSTANCE_GPU_MAP[instance_type]
            string = AwsUtils._format_price_string(
                instance_type, round(price, 2), f"{num_gpus} x {gpu_type}", cpus, mem
            )
            print(string)
        print()

    @staticmethod
    def ls_spot_gpu_prices(dry_run=False) -> None:
        if dry_run:
            return

        ec2 = CONFIG.get_session().client("ec2")

        string = AwsUtils._format_price_string(
            "InstanceType", "Price ($/hour)", "GPUs", "CPUs", "Mem (Gb)"
        )
        print()
        nprint(string, style="bold")

        for instance_type in AwsUtils._instance_types():
            response = ec2.describe_spot_price_history(
                InstanceTypes=[instance_type],
                Filters=[{"Name": "product-description", "Values": ["Linux/UNIX"]}],
            )

            price = float(response["SpotPriceHistory"][0]["SpotPrice"])

            num_gpus, gpu_type, mem, cpus = INSTANCE_GPU_MAP[instance_type]
            string = AwsUtils._format_price_string(
                instance_type, round(price, 2), f"{num_gpus} x {gpu_type}", cpus, mem
            )
            print(string)
        print()

    @staticmethod
    def _instance_types() -> Generator[str, None, None]:
        """Yield all relevant EC2 instance types in region CONFIG.region_name"""

        describe_args = {}
        client = CONFIG.get_session().client("ec2")

        def instance_type_generator():
            while True:
                describe_result = client.describe_instance_types(**describe_args)
                yield from (i["InstanceType"] for i in describe_result["InstanceTypes"])
                if "NextToken" not in describe_result:
                    break
                describe_args["NextToken"] = describe_result["NextToken"]

        return (
            inst
            for inst in sorted(instance_type_generator())
            if inst.startswith(("p2", "p3", "p4", "g4d"))
        )

    @staticmethod
    def _format_price_string(instance_type, price, gpus, cpus, mem) -> str:
        string = "\t{0: <16} {1: <15} {2: <10} {3: <5} {4:<7}".format(
            instance_type, price, gpus, cpus, mem
        )
        return string

    @staticmethod
    def spending(qty: int, timescale: str, dry_run=False) -> None:
        today = date.today()
        time_format = "%Y-%m-%d"

        if timescale == "months":
            start_date = (today - relativedelta(months=qty)).replace(day=1)
            start = start_date.strftime(time_format)
            end = today.strftime(time_format)
            granularity = "monthly"
        elif timescale == "days":
            start_date = today - relativedelta(days=qty)
            start = start_date.strftime(time_format)
            end = today.strftime(time_format)
            granularity = "daily"
        else:
            raise ValueError("Timescale must be 'daily' or 'monthly'.")

        services = [
            "Amazon Elastic Compute Cloud - Compute",
            "EC2 - Other",
            "Amazon Simple Storage Service",
        ]

        client = CONFIG.get_session().client("ce")
        results = client.get_cost_and_usage(
            TimePeriod={"End": end, "Start": start},
            Granularity=granularity.upper(),
            Filter={
                "And": [
                    {
                        "Not": {
                            "Dimensions": {
                                "Key": "RECORD_TYPE",
                                "Values": ["Credit", "Refund"],
                            }
                        }
                    },
                    {"Dimensions": {"Key": "SERVICE", "Values": services}},
                ]
            },
            Metrics=["UnblendedCost"],
            GroupBy=[{"Type": "DIMENSION", "Key": "SERVICE"}],
        )["ResultsByTime"]

        table = []
        print()
        for interval in results:
            period_start_date = datetime.strptime(
                interval["TimePeriod"]["Start"], "%Y-%m-%d"
            )
            if granularity == "monthly":
                period_string = period_start_date.strftime("%b %Y")
            else:
                period_string = period_start_date.strftime("%d %b")

            groups = interval["Groups"]
            if not groups:
                table.append([period_string, 0, 0])
            else:
                ec2_cost = 0
                s3_cost = 0
                for group in groups:
                    group_cost = float(group["Metrics"]["UnblendedCost"]["Amount"])
                    if group["Keys"] == ["Amazon Elastic Compute Cloud - Compute"]:
                        ec2_cost += group_cost
                    elif group["Keys"] == ["EC2 - Other"]:
                        ec2_cost += group_cost
                    elif group["Keys"] == ["Amazon Simple Storage Service"]:
                        s3_cost += group_cost

                table.append([period_string, ec2_cost, s3_cost])

        def row_string(x):
            string = f"\t{x[0]:>10}"
            for xi in x[1:]:
                if type(xi) == float:
                    string += f" {xi:>10.2f}"
                else:
                    string += f" {xi:>10}"
            return string

        print(f"\tSpending for region {CONFIG.region_name}:")
        print()
        print(row_string(["", "EC2", "S3"]))
        for row in table:
            # round before passing to row_string
            print(row_string(row))

        ec2_total = sum([row[1] for row in table])
        s3_total = sum([row[2] for row in table])
        print("\t" + "-" * 32)
        print(row_string(["Total", ec2_total, s3_total]))
        print()



import pytest
from click.testing import CliRunner

from nimbo import CONFIG
from nimbo.core.cloud_provider.provider_impl.aws.aws_provider import AwsProvider
from nimbo.core.config import RequiredCase
from nimbo.tests.aws.utils import isolated_filesystem


@pytest.fixture
def runner():
    return ""


@isolated_filesystem(RequiredCase.MINIMAL)
def test_get_image_id(runner: CliRunner):
    reference_image = "ubuntu18-latest-drivers"
    reference_region = "eu-west-1"

    CONFIG.image = reference_image
    CONFIG.region_name = reference_region

    AwsProvider._get_image_id()

    CONFIG.image = "awdghiuadgwui"
    with pytest.raises(ValueError):
        AwsProvider._get_image_id()

    CONFIG.image = reference_image
    CONFIG.region_name = "us-east-2"
    AwsProvider._get_image_id()

    with pytest.raises(ValueError):
        CONFIG.region_name = "us-et-2"
        AwsProvider._get_image_id()

    CONFIG.image = "ami-198571934781039"
    CONFIG.region_name = reference_region
    AwsProvider._get_image_id()



import abc
import os
import random
import string
from typing import Any, Dict, Optional, Union

import pydantic

from nimbo.core.config import (
    AwsConfig,
    CloudProvider,
    GcpConfig,
    RequiredCase,
    yaml_loader,
)

CONDA_ENV = "env.yml"
ASSETS_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def _random_string_of_len(length: int) -> str:
    return "".join(random.choice(string.ascii_letters) for _ in range(length))


S3_DATASETS_PATH = f"s3://nimbo-test-bucket/{_random_string_of_len(20)}"
S3_RESULTS_PATH = f"s3://nimbo-test-bucket/{_random_string_of_len(20)}"


class CommonTestConfigMixin(abc.ABC, pydantic.BaseModel):
    # Can't use pydantic.PrivateAttr here as pydantic complains about overlap later
    # user_id and _initial_state fields defined here to prevent undefined warnings
    user_id: Optional[str] = None
    _initial_state: Dict[str, Any] = {}

    def save_initial_state(self):
        self._initial_state = self.dict()

    def reset_required_config(self) -> None:
        if len(self._initial_state) <= 1:
            raise ValueError(
                "You must run save_initial_state to use reset_required_config"
            )

        # user_id is set after save_initial_state
        if not self._initial_state["user_id"]:
            self._initial_state["user_id"] = self.user_id

        for key, value in self._initial_state.items():
            setattr(self, key, value)

    @abc.abstractmethod
    def inject_required_config(self, *cases: RequiredCase) -> None:
        ...


class AwsTestConfig(CommonTestConfigMixin, AwsConfig):
    _initial_state: Dict[str, Any] = pydantic.PrivateAttr(default_factory=dict)

    def inject_required_config(self, *cases: RequiredCase) -> None:
        cases = RequiredCase.decompose(*cases)

        if RequiredCase.NONE in cases:
            self.telemetry = False
        if RequiredCase.MINIMAL in cases:
            self.region_name = "eu-west-1"
        if RequiredCase.STORAGE in cases:
            self.local_datasets_path = "random_datasets"
            self.local_results_path = "random_results"
            self.s3_datasets_path = S3_DATASETS_PATH
            self.s3_results_path = S3_RESULTS_PATH
        if RequiredCase.INSTANCE in cases:
            self.image = "ubuntu18-latest-drivers"
            self.disk_size = 64
            self.instance_type = "p2.xlarge"
            self.instance_key = self._find_instance_key(ASSETS_PATH, self.region_name)
        if RequiredCase.JOB in cases:
            self.conda_env = CONDA_ENV

    @staticmethod
    def _find_instance_key(directory: str, region_name: str) -> str:
        # Instance keys are supposed to be prefixed with region name for testing
        for file in os.listdir(directory):
            if file.startswith(region_name) and file.endswith(".pem"):
                return file


class GcpTestConfig(CommonTestConfigMixin, GcpConfig):
    _initial_state: Dict[str, Any] = pydantic.PrivateAttr(default_factory=dict)

    def inject_required_config(self, *cases: RequiredCase) -> None:
        pass


def make_config(config_path: str) -> Union[AwsTestConfig, GcpTestConfig]:
    raw_config = yaml_loader.from_file(os.path.join(ASSETS_PATH, config_path))
    raw_config["config_path"] = config_path

    cloud_provider = CloudProvider(raw_config["cloud_provider"].upper())

    if cloud_provider == CloudProvider.AWS:
        config = AwsTestConfig(**raw_config)
    else:
        config = GcpTestConfig(**raw_config)

    config.save_initial_state()

    return config



import os

import pytest
from click.testing import CliRunner

from nimbo import CONFIG
from nimbo.core.cloud_provider.provider_impl.aws.services.aws_instance import (
    AwsInstance,
)
from nimbo.core.config import RequiredCase
from nimbo.core.constants import FULL_REGION_NAMES
from nimbo.main import cli
from nimbo.tests.aws.utils import isolated_filesystem


@pytest.fixture
def runner():
    return ""


@isolated_filesystem(RequiredCase.INSTANCE, RequiredCase.STORAGE)
def test_test_access(runner: CliRunner):
    result = runner.invoke(cli, "test-access", catch_exceptions=False)
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.JOB)
def test_run_no_code(runner: CliRunner):
    result = runner.invoke(cli, "rm-all-instances", input="y", catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(cli, "run 'python --version'", catch_exceptions=False)
    assert result.exit_code == 0
    result = runner.invoke(cli, "rm-all-instances", input="y", catch_exceptions=False)
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.JOB)
def test_launch(runner: CliRunner):
    dst = os.getcwd()
    instance_keys = [p for p in os.listdir(dst) if p[-4:] == ".pem"]

    for instance_key in instance_keys:
        for region_name in FULL_REGION_NAMES.keys():
            if instance_key.startswith(region_name):
                CONFIG.region_name = region_name
                break
        else:
            raise Exception(
                f"Instance key {instance_key} does not begin with a valid region name"
            )

        CONFIG.instance_key = instance_key
        response = AwsInstance.run("_nimbo_launch", dry_run=False)

        assert response["message"] == "_nimbo_launch_success"
        instance_id = response["instance_id"]

        result = runner.invoke(
            cli, f"rm-instance {instance_id}", catch_exceptions=False
        )
        assert result.exit_code == 0


@isolated_filesystem(RequiredCase.JOB)
def test_spot_launch(runner: CliRunner):
    CONFIG.spot = True
    response = AwsInstance.run("_nimbo_launch", dry_run=False)

    assert response["message"] == "_nimbo_launch_success"
    instance_id = response["instance_id"]

    result = runner.invoke(cli, f"rm-instance {instance_id}", catch_exceptions=False)
    assert result.exit_code == 0


@isolated_filesystem(RequiredCase.JOB)
def test_notebook(runner: CliRunner):
    CONFIG.spot = True
    response = AwsInstance.run("_nimbo_notebook", dry_run=False)

    assert response["message"] == "_nimbo_notebook_success"
    instance_id = response["instance_id"]

    result = runner.invoke(cli, f"rm-instance {instance_id}", catch_exceptions=False)
    assert result.exit_code == 0



import json
from datetime import datetime

import requests

from nimbo import CONFIG


# noinspection PyBroadException
def record_event(cmd):
    if not CONFIG.telemetry:
        return
    else:
        if not CONFIG.user_id:
            CONFIG.get_session()

    now = datetime.now()
    date_time = now.strftime("%Y-%m-%d-%H-%M-%S")

    data = {
        "user_id": CONFIG.user_id,
        "user_arn": CONFIG.user_arn,
        "cmd": cmd,
        "date": date_time,
    }
    try:
        requests.post(CONFIG.telemetry_url, data=json.dumps(data), timeout=2)
    except BaseException:
        pass



import enum
import os
import sys
from typing import Optional

import boto3
import botocore
import botocore.session
import pydantic

from nimbo.core.config.common_config import BaseConfig, RequiredCase
from nimbo.core.constants import FULL_REGION_NAMES


class _DiskType(str, enum.Enum):
    STANDARD = "standard"
    IO1 = "io1"
    IO2 = "io2"
    GP2 = "gp2"
    SC1 = "sc1"
    ST1 = "st1"
    GP3 = "gp3"


class _Encryption(str, enum.Enum):
    AES256 = "AES256"
    AWSKMS = "aws:kms"


class AwsConfig(BaseConfig):
    aws_profile: Optional[str] = None
    region_name: Optional[str] = None

    s3_datasets_path: Optional[str] = None
    s3_results_path: Optional[str] = None
    encryption: _Encryption = None

    instance_type: Optional[str] = None
    image: str = "ubuntu18-latest-drivers"
    disk_size: Optional[int] = None
    disk_iops: pydantic.conint(ge=0) = None
    disk_type: _DiskType = _DiskType.GP2
    spot: bool = False
    spot_duration: pydantic.conint(ge=60, le=360, multiple_of=60) = None
    security_group: Optional[str] = None
    instance_key: Optional[str] = None
    role: Optional[str] = None

    # The following are defined internally
    user_arn: Optional[str] = None

    def get_session(self) -> boto3.Session:
        session = boto3.Session(
            profile_name=self.aws_profile, region_name=self.region_name
        )

        caller_identity = session.client("sts").get_caller_identity()
        self.user_id = caller_identity["UserId"]
        self.user_arn = caller_identity["Arn"]

        return session

    def assert_required_config_exists(self, *cases: RequiredCase) -> None:
        """ Designed to be used with the assert_required_config annotation """

        cases = RequiredCase.decompose(*cases)

        if len(cases) == 1 and RequiredCase.NONE in cases:
            return
        elif not os.path.isfile(self.config_path):
            raise FileNotFoundError(
                f"Nimbo configuration file '{self.config_path}' not found.\n"
                "Run 'nimbo generate-config' to create the default config file."
            )

        required_config = {}

        if RequiredCase.MINIMAL in cases:
            required_config["cloud_provider"] = self.cloud_provider
            required_config["aws_profile"] = self.aws_profile
            required_config["region_name"] = self.region_name
        if RequiredCase.STORAGE in cases:
            required_config["local_results_path"] = self.local_results_path
            required_config["local_datasets_path"] = self.local_datasets_path
            required_config["s3_results_path"] = self.s3_results_path
            required_config["s3_datasets_path"] = self.s3_datasets_path
        if RequiredCase.INSTANCE in cases:
            required_config["instance_type"] = self.instance_type
            required_config["disk_size"] = self.disk_size
            required_config["instance_key"] = self.instance_key
            required_config["security_group"] = self.security_group
            required_config["role"] = self.role
        if RequiredCase.JOB in cases:
            required_config["conda_env"] = self.conda_env

        unspecified = [key for key, value in required_config.items() if not value]
        if unspecified:
            raise AssertionError(
                f"For running this command '{', '.join(unspecified)}' should"
                f" be specified in {self.config_path}"
            )

        bad_fields = {}

        if RequiredCase.MINIMAL in cases:
            bad_fields["aws_profile"] = self._aws_profile_exists()
            bad_fields["region_name"] = self._region_name_valid()
        if RequiredCase.STORAGE in cases:
            bad_fields["local_results_path"] = self._local_results_not_outside_project()
            bad_fields[
                "local_datasets_path"
            ] = self._local_datasets_not_outside_project()
        if RequiredCase.INSTANCE in cases:
            bad_fields["instance_key"] = self._instance_key_valid()
            bad_fields["disk_iops"] = self._disk_iops_specified_when_needed()
        if RequiredCase.JOB in cases:
            bad_fields["conda_env"] = self._conda_env_valid()

        bad_fields = [(key, error) for key, error in bad_fields.items() if error]

        if bad_fields:
            print(
                f"{len(bad_fields)} error{'' if len(bad_fields) == 1 else 's'} "
                f"in {self.config_path}\n"
            )
            for key, error in bad_fields:
                print(key)
                print(f"  {error}")
            sys.exit(1)

    def _aws_profile_exists(self) -> Optional[str]:
        if self.aws_profile not in botocore.session.Session().available_profiles:
            return f"AWS Profile '{self.aws_profile}' could not be found"

    def _conda_env_valid(self) -> Optional[str]:
        if os.path.isabs(self.conda_env):
            return "conda_env should be a relative path"
        if ".." in self.conda_env:
            return "conda_env should not be outside of the project directory"
        if not os.path.isfile(self.conda_env):
            return f"file '{self.conda_env}' does not exist in the project directory"

    def _instance_key_valid(self) -> Optional[str]:
        if not os.path.isfile(self.instance_key):
            return f"file '{self.instance_key}' does not exist"

        permission = str(oct(os.stat(self.instance_key).st_mode))[-3:]
        if permission[1:] != "00":
            return (
                f"run 'chmod 400 {self.instance_key}' so that only you can read the key"
            )

    def _region_name_valid(self) -> Optional[str]:
        region_names = FULL_REGION_NAMES.keys()
        if self.region_name not in region_names:
            return (
                f"unknown region '{self.region_name}', "
                f"expected to be one of {', '.join(region_names)}"
            )

    def _disk_iops_specified_when_needed(self) -> Optional[str]:
        if self.disk_type in [_DiskType.IO1, _DiskType.IO2] and not self.disk_iops:
            return (
                "for disk types io1 or io2, the 'disk_iops' parameter has "
                "to be specified.\nPlease visit "
                "https://docs.nimbo.sh/nimbo-config-file-options for more details."
            )



from pprint import pprint

import boto3
import firebase_admin
from firebase_admin import credentials, db

from nimbo import CONFIG
from nimbo.core.constants import NIMBO_ROOT

cred = credentials.Certificate(
    NIMBO_ROOT + "/ami/credentials/nimboami-firebase-adminsdk-yxyfy-6d01f1d35f.json"
)
firebase_admin.initialize_app(
    cred, options={"databaseURL": "https://nimboami-default-rtdb.firebaseio.com"}
)
image_db = db.reference("images")

image_catalog = {}
for dest_region in CONFIG.full_region_names.keys():
    print(dest_region)
    session = boto3.Session(profile_name="nimbo", region_name=dest_region)
    ec2 = session.client("ec2")
    response = ec2.describe_images(
        Filters=[
            {"Name": "tag:CreatedBy", "Values": ["nimbo"]},
            {"Name": "tag:Type", "Values": ["production"]},
        ]
    )
    images = response["Images"]
    image_catalog_per_region = {img["Name"]: img["ImageId"] for img in images}
    image_catalog[dest_region] = image_catalog_per_region.copy()
    pprint(image_catalog_per_region)

image_db.set(image_catalog)



from nimbo.core.cloud_provider.provider.services.permissions import Permissions


class GcpPermissions(Permissions):
    @staticmethod
    def mk_instance_key(dry_run=False) -> None:
        pass

    @staticmethod
    def allow_ingress_current_ip(target: str, dry_run=False) -> None:
        ...

    @staticmethod
    def setup(profile: str, no_s3_access=False) -> None:
        pass

    @staticmethod
    def add_user(profile: str, username: str) -> None:
        pass



from nimbo.core.cloud_provider.provider.services.utils import Utils


class GcpUtils(Utils):
    @staticmethod
    def ls_gpu_prices(dry_run=False) -> None:
        pass

    @staticmethod
    def ls_spot_gpu_prices(dry_run=False) -> None:
        pass

    @staticmethod
    def spending(qty: int, timescale: str, dry_run=False) -> None:
        pass



import abc


class Storage(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def push(folder: str, delete=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def pull(folder: str, delete=False) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def ls_bucket(path: str) -> None:
        ...

    @staticmethod
    @abc.abstractmethod
    def mk_bucket(bucket_name: str, dry_run=False) -> None:
        ...



import functools
import re
import sys
import typing as t

import pkg_resources
import pydantic

import nimbo.core.config
import nimbo.tests.aws.config
from nimbo.core.config import RequiredCase
from nimbo.core.config.aws_config import AwsConfig
from nimbo.core.config.common_config import CloudProvider
from nimbo.core.config.gcp_config import GcpConfig
from nimbo.core.constants import IS_TEST_ENV
from nimbo.core.print import nprint

version = pkg_resources.get_distribution("nimbo").version


CONFIG: t.Optional[t.Union[AwsConfig, GcpConfig]] = None
_CLOUD = None


def set_config(config_factory, config_path):
    global CONFIG, _CLOUD

    try:
        CONFIG = config_factory(config_path)
    except pydantic.error_wrappers.ValidationError as e:
        e_msg = str(e)
        e_num = len(e.errors())
        title_end = e_msg.index("\n", 1)
        new_title = (
            f"{e_num} error{'' if e_num == 1 else 's'} in "
            f"{config_path if config_path else 'nimbo-config.yml'}\n"
        )
        print(new_title + re.sub(r"\(type=.*\)", "", e_msg[title_end:]))
        sys.exit(1)

    # Both of these imports depend on this file
    from nimbo.core.cloud_provider.provider_impl.aws.aws_provider import AwsProvider
    from nimbo.core.cloud_provider.provider_impl.gcp.gcp_provider import GcpProvider

    if CONFIG.cloud_provider == CloudProvider.AWS:
        _CLOUD = AwsProvider()
    else:
        _CLOUD = GcpProvider()


if IS_TEST_ENV:
    set_config(nimbo.tests.aws.config.make_config, "nimbo-config.yml")


def assert_required_config(*cases: RequiredCase):
    """
    Decorator for ensuring that required config is present
    """

    def decorator(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            try:
                CONFIG.assert_required_config_exists(*cases)
                return func(*args, **kwargs)
            except AssertionError as e:
                nprint(e, style="error")
                sys.exit(1)
            except FileNotFoundError as e:
                # Happens when nimbo config file is not found
                nprint(e, style="error")
                sys.exit(1)

        return decorated

    return decorator


def cloud_context(func):
    """
    Decorator for injecting a key-value argument cloud of type AwsProvider|GcpProvider
    """

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        kwargs["cloud"] = _CLOUD
        return func(*args, **kwargs)

    return decorated



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nimbo",  # Replace with your own username
    version="0.3.1",
    author="NimboSH, Ltd.",
    author_email="support@nimbo.sh",
    description="Run machine learning jobs on AWS with a single command.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nimbo-sh/nimbo",
    project_urls={"Bug Tracker": "https://github.com/nimbo-sh/nimbo/issues"},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    package_data={"nimbo": ["scripts/*.sh"]},
    include_package_data=True,
    entry_points={"console_scripts": ["nimbo=nimbo.main:cli"]},
    python_requires=">=3.6",
    # Colorama version specified to resolve conflicts with awcli, click, rich
    install_requires=[
        "colorama==0.4.3",
        "requests>=2.25",
        "click>=7.0",
        "pydantic>=1.7.0",
        "pyyaml>=5.3.0",
        "boto3>=1.17",
        "awscli>=1.19<2.0",
        "rich>=10.1.0",
        "python-dateutil>=2.8.0",
    ],
)



import sys

import boto3

from nimbo import CONFIG

image_id = sys.argv[1]

base_region = "eu-west-1"
session = boto3.Session(profile_name="nimbo", region_name=base_region)
ec2 = session.client("ec2")
response = ec2.describe_images(
    Filters=[
        {"Name": "tag:CreatedBy", "Values": ["nimbo"]},
        {"Name": "tag:Type", "Values": ["production"]},
    ],
    ImageIds=[image_id],
)
image = response["Images"][0]

for dest_region in CONFIG.full_region_names.keys():
    if dest_region == base_region:
        continue
    print(dest_region)
    session = boto3.Session(profile_name="nimbo", region_name=dest_region)
    ec2 = session.client("ec2")

    response = ec2.copy_image(
        ClientToken=image_id,
        Name=image["Name"],
        SourceImageId=image_id,
        SourceRegion=base_region,
        Description=image["Description"],
    )
    new_image_id = response["ImageId"]
    ec2.create_tags(
        Resources=[new_image_id],
        Tags=[
            {"Key": "CreatedBy", "Value": "nimbo"},
            {"Key": "Type", "Value": "production"},
        ],
    )



from rich.console import Console
from rich.theme import Theme

_console = Console(
    theme=Theme(
        {
            "repr.number": "",
            "repr.str": "",
            "repr.ellipsis": "",
            "repr.eui48": "",
            "repr.eui64": "",
            "repr.ipv4": "",
            "repr.ipv6": "",
            "repr.filename": "",
            "repr.path": "",
            "error": "bold red",
            "warning": "bold magenta",
        }
    )
)
nprint = _console.print


def nprint_header(x) -> None:
    nprint(f"[blue]==>[/blue] {x}", style="bold")
